/*
  dmp/device.cpp

  Copyright (c) 2021 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#include "device.h"

namespace ACN::DMP {

/**
 * @brief Device::Device
 */
Device::Device()
{

}

/**
 * @brief Device::~Device
 */
Device::~Device()
{
  delete properties;
}


/**
 * @brief Device::DmpReceiver
 * @param block
 */
void Device::dmpReceiver(PDU::Block<DMP::Pdu> block)
{
  /// \cite DMP 5.1.2 Relative Addresses
  /// Relative addresses are used to specify an address as an offset from the
  /// most recently used address within the same PDU Block.
  last_address_used = 0;

  for(const auto& dmp : *block.pdu)
    {
      switch (dmp->vector()) {
        case GET_PROPERTY:
          dmpGet(dmp);
          break;
        case SET_PROPERTY:
          dmpSet(dmp);
          break;
        default:
          break;
        }
    }
}


/**
 * @brief Device::dmpGet
 * @param dmp
 */
void Device::dmpGet(PDU::Message<DMP::Pdu> dmp)
{
  (void)dmp;
}


/**
 * @brief Device::dmpSet
 * @param dmp
 */
void Device::dmpSet(PDU::Message<DMP::Pdu> dmp)
{
  auto type = std::static_pointer_cast<address_type>(dmp->header());
  auto set_data = std::static_pointer_cast<address_pair_list>(dmp->data());

  for(const auto& [range, data] : set_data->properties)
    {
      /// \cite DMP 5.1.2 Relative Addresses
      /// Relative addresses are used to specify an address as an offset
      /// from the most recently used address within the same PDU Block.
      uint32_t address = range.address;
      if (type->relative)
        address += last_address_used;

      /// \todo do something with DMP properties

      /// \cite DMP 5.1.2 Relative Addresses
      /// If a command addresses a range of properties then a subsequent
      /// relative address shall be relative to the last property in that range.
      last_address_used = address;
    }
}


/**
 * @brief Device::getProperty
 * @param address
 * @return
 */
uint32_t Device::getProperty(const unsigned int address) const
{
  if (!properties)
    return 0;
  if (properties->size() < address)
    return 0;
  return properties->at(address);
}


/**
 * @brief Device::setProperty
 * @param address
 * @param value
 */
void Device::setProperty(const unsigned int address, const int value)
{
  if (!properties)
    return;
  if (properties->size() < address)
    return;
  properties->at(address) =  value;
}

} // namespace ACN::DMP
