/*
  universe.cpp

  Copyright (c) 2020 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#include "universe.h"

namespace DMX {

/**
 * @brief Universe::Universe
 * @param timeout_period
 */
Universe::Universe(int timeout_period)
  : null_start_data({0})
  , rx_timeout_period_(timeout_period)
{
}


/**
 * @brief Universe::~Universe
 */
Universe::~Universe()
{
}


/**
 * @brief Universe::slot
 * @param address
 * @return
 */
uint8_t Universe::slot(const uint16_t address) const
{
  if (address == 0)
    return 0;
  if (address > null_start_data.size() - 1)
    return 0;

  null_start_mutex.lock();
  uint8_t val = null_start_data[address];
  null_start_mutex.unlock();

  return val;
}


/**
 * @brief Universe::rxRate Rate (in Hz) at which the universe is being recv'd.
 * @return Hz
 *
 *  Calculated as the rolling mean of (the number of frames in
 *  the last rx_timeout_peroid ms) / rx_timeout_period seconds.
 */
double Universe::rxRate()
{
    rx_timeout_();
    //         updates     per      second
    return rx_times_.size() / (rx_timeout_period_ / 1000.0);
}


/**
 * @brief Universe::setData Accept new data from receiver
 * @param data varaibly sized data, beginning with a start code.
 *
 * The only setData method used to establish rx rate.
 */
void Universe::setData(const std::vector<uint8_t>& data)
{
  // no data?
  if (data.empty())
    return;

  // Alternate Start Code
  if (data.front() != E111_NULL_START)
    return altSCdata(data);

  // too many slots!
  if (data.size() > null_start_data.size())
    return;

  // copy data
  null_start_mutex.lock();
  if (data.size() < null_start_data.size())
    null_start_data.fill(0);
  std::copy(data.begin(), data.end(), null_start_data.begin());
  null_start_mutex.unlock();

  // update rx times
  rx_timeout_(true);

  // notify callbacks
  for (const auto &cb : callbacks_)
    cb(this);
}


/**
 * @brief Universe::altSCdata
 * @param data
 *
 * \attention re-impliment this method to rx custom alternate start code data.
 */
void Universe::altSCdata(const std::vector<uint8_t> & data)
{
  switch (data.front()) {
    case E111_ASC_TEXT_ASCII:
      break;
    case E111_ASC_TEST:
      break;
    case E111_ASC_TEXT_UTF8:
      break;
    case E111_ASC_MANUFACTURER:
      break;
    case E111_ASC_SIP:
      break;
    }
}


/**
 * @brief Universe::setValue
 * @param address
 * @param value
 */
void Universe::setValue(const uint16_t address, const uint8_t value)
{
    if (address == 0)
      return;
    if (address > null_start_data.size() - 1)
      return;

    null_start_mutex.lock();
    null_start_data[address] = value;
    null_start_mutex.unlock();
}


/**
 * @brief Universe::setValue
 * @param start
 * @param footprint
 * @param data
 */
void Universe::setValue(const uint16_t start, const uint16_t footprint,
                        const uint8_t* data)
{
    if (start == 0)
      return;
    if (start + footprint > null_start_data.size() - 1)
      return;

    null_start_mutex.lock();
    std::copy(data, data + footprint, null_start_data.begin() + start);
    null_start_mutex.unlock();
}


/**
 * @brief Universe::onData
 * @param callback
 */
void Universe::onData(const DataHandler callback)
{
  callbacks_.push_back(callback);
}


/**
 * @brief Universe::rx_timeout_
 * @param add_now
 */
void Universe::rx_timeout_(bool add_now)
{
  auto now = std::chrono::system_clock::now();
  auto elapsed = [] (auto& a, auto& b) {
      return std::chrono::duration_cast<std::chrono::milliseconds>(a - b);
    };

  if (add_now)
    rx_times_.push(now);

  while (rx_times_.size() > 0 &&
         elapsed(now, rx_times_.front()).count() > rx_timeout_period_)
    rx_times_.pop();
}

} // namespace DMX
