/*
  universe.h

  Copyright (c) 2020 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/
#pragma once

#include "dmx.h"

#include <chrono>
#include <cstdint>
#include <functional>
#include <mutex>
#include <queue>
#include <vector>

namespace DMX {

class Universe; // forward declare the Univserse class
using DataHandler = std::function<void(Universe *)>;
using DimmerData  = std::array<uint8_t, 513>;

/**
 * @brief The Universe class
 *
 *  ANSI E1.11 describes may things about moving data over serial EIA-485-A
 *  links.  This class encapselates that data, regardless of how it was
 *  transmitted.
 */
class Universe {
  public:
    Universe (int timeout_period = E111_DATA_LOSS_TIMEOUT);
    virtual ~Universe ();

    virtual uint8_t slot (const uint16_t) const;
    virtual double rxRate();

    virtual void setValue (const uint16_t address, const uint8_t value);
    virtual void setValue (const uint16_t start, const uint16_t footprint,
                           const uint8_t* data);

    void setData (const std::vector<uint8_t> &);
    void onData (const DataHandler);

    virtual void altSCdata(const std::vector<uint8_t> &);

  protected:
    DimmerData null_start_data;           //!< NULL Start Code data
    mutable std::mutex null_start_mutex;  //!< memory protect Null Start data

    std::vector<DataHandler> callbacks_;  //!< list of calback functions

  private:
    std::queue<std::chrono::time_point<std::chrono::system_clock>> rx_times_;

    void rx_timeout_(bool add_now = false);
    const int rx_timeout_period_;
};

} // namespace DMX
