/*
  ept.h

  Copyright (c) 2021 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/
#pragma once

#include "rdmnet/pdu.h"
#include "uuid/uuid.h"

namespace RDMnet::EPT {

/**
 * @brief 1.2.4 Extensible Packet Transport (EPT)
 *
 * EPT allows non-RDM data to be transmitted using a Broker. It uses a simple
 * message format which addresses Components solely by their CID, and allows
 * manufacturer-defined data to be transmitted between Components free from the
 * message format and behavioral restrictions of RDM.
 */


/**
 * @brief The EPT::Pdu class
 */
class Pdu
  : public RDMnet::Pdu
{
public:
  Pdu();
  void iStream(ACN::PDU::Stream) override;
};


/**
 * @brief The ept_header struct
 */
struct ept_header
    : ACN::PDU::pdu_header
{
  UUID::uuid destination_cid; //!< CID of the destination component

  size_t streamSize() const override { return UUID_LENGTH; }
  void   iStream(ACN::PDU::Stream s) override { *s >> destination_cid; }
  void   oStream(ACN::PDU::Stream s) const override { *s << destination_cid; }
};


namespace DATA {

/**
 * @brief The EPT::DATA::Pdu class
 */
class Pdu
  : public RDMnet::Pdu
{
public:
  Pdu();
  void iStream(ACN::PDU::Stream) override;
};


/**
 * @brief The data_data struct
 */
struct data_data
    : ACN::PDU::pdu_data
{
  std::vector<uint8_t> data; //!< unprocessed data array

  size_t streamSize() const override { return data.size(); }
  void   iStream(ACN::PDU::Stream) override;
  void   oStream(ACN::PDU::Stream) const override;
};

} // namespace DATA

namespace STATUS {

/**
 * @brief The EPT::STATUS::Pdu class
 */
class Pdu
  : public RDMnet::Pdu
{
public:
  Pdu();
  void iStream(ACN::PDU::Stream) override;
};

} // namespace STATUS

} // namespace RDMnet::EPT
