/*
  rlp/udp.cpp

  Copyright (c) 2020 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#include "udp.h"

#include <cstring>

namespace ACN::RLP::UDP {

/**
 * @brief preamble::preamble
 * @param filled
 */
transport::transport(bool filled)
  : RLP::message_transport()
{
  if (!filled)
    return;

  length = sizeof(length) + sizeof(postamble_size) + sizeof(acn_id);
  postamble_size = 0; //!< 3. The postamble field shall be empty.
  memcpy(acn_id, ACN_PACKET_IDENTIFIER, sizeof(acn_id));
}


/**
 * @brief preamble_t::iStream
 * @param stream
 */
void transport::iStream(PDU::Stream stream)
{
  if (!stream->good())
    return;
  /**
   *  2. Preamble Format:
   *  The preamble size includes both size fields so the
   *  minimum value for preamble size is 16 (octets).
   */
  *stream >> length;
  if (length < PREAMBLE_MINIMUM_SIZE)
    return stream->setstate(std::ios_base::failbit);

  *stream >> postamble_size;

  /**
   *  2. Preamble Format:
   *  The ACN Packet Identifier shall be the text string “ASC-E1.17\0\0\0”
   *  encoded in [ASCII].
   */
  stream->read(acn_id, sizeof(acn_id));
  if (stream->gcount() != sizeof(acn_id))
    return stream->setstate(std::ios_base::failbit);
  if (memcmp(acn_id, ACN_PACKET_IDENTIFIER, sizeof(acn_id))) // memcmp returns 0 if matched
    return stream->setstate(std::ios_base::failbit);


  /**
   * Implementations shall compute the size and position of the PDU block from
   * the preamble size and postamble size provided. ... ignoring any extra
   * octets in the preamble or postamble.
   */
  if (length > PREAMBLE_MINIMUM_SIZE)
    stream->ignore((length - PREAMBLE_MINIMUM_SIZE));

  RLP::message_transport::iStream(stream);
}


/**
 * @brief preamble_t::oStream
 * @param stream
 */
void transport::oStream(PDU::Stream stream) const
{
  *stream << length;
  *stream << postamble_size;
  stream->write(acn_id, sizeof(acn_id));
  RLP::message_transport::oStream(stream);
}

} // ACN::RLP::UDP
