/*
  mergeproxyuniverse.cpp

  Copyright (c) 2020 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#include "arbitratinguniverse.h"

namespace sACN {

/**
 * @brief MergeProxyUniverse::MergeProxyUniverse
 */
ArbitratingUniverse::ArbitratingUniverse()
      : sACN::Universe()
      , expectedUniverse(0)
{
}


/**
 * @brief MergeProxyUniverse::~MergeProxyUniverse
 */
ArbitratingUniverse::~ArbitratingUniverse()
{
  for (auto& [_, universe] : sources_)
    delete universe;
}


/**
 * @brief MergeProxyUniverse::sources
 * @return
 */
const std::vector<DATA::data_header> ArbitratingUniverse::sources() const
{
  std::vector<DATA::data_header> keys;
  for (const auto& [key, _] : sources_)
    keys.push_back(key);
  return keys;
}


/**
 * @brief MergeProxyUniverse::sourceUniverse
 * @param src
 * @return
 */
Universe* ArbitratingUniverse::sourceUniverse(const DATA::data_header &src)
{
  if (!hasSourceUniverse(src))
    newProvenance_(src);
  return sources_.at(src);
}


/**
 * @brief MergeProxyUniverse::hasSourceUniverse
 * @param src
 * @return
 */
bool ArbitratingUniverse::hasSourceUniverse(const DATA::data_header& src) const
{
  if (sources_.count(src))
    return true;
  return false;
}


/**
 * @brief MergeProxyUniverse::hassources
 * @return
 */
bool ArbitratingUniverse::hasSources() const
{
  if (sources_.size())
    return true;
  return false;
}


/**
 * @brief MergeProxyUniverse::deleteSourceUniverse
 * @param src
 */
void ArbitratingUniverse::deleteSourceUniverse(const DATA::data_header& src)
{
  if (sources_.count(src))
    {
      delete sources_.at(src);
      sources_.erase(src);
    }
}


/**
 * @brief MergeProxyUniverse::addSourceUniverse
 * @param universe
 */
void ArbitratingUniverse::addSourceUniverse(Universe* universe)
{
  sources_.at(*universe->provenance()) = universe;
}


/**
 * @brief MergeProxyUniverse::dataChangedNotifier
 * @param dmx
 */
void ArbitratingUniverse::dataChangedNotifier(DMX::Universe* dmx)
{
  auto sacn = static_cast<sACN::Universe*>(dmx);
  auto universe = dominant_();
  if (!universe)
    return;
  if (sacn->provenance() == universe->provenance())
    for (const auto &cb : callbacks_)
      cb(this);
}


/**
 * @brief MergeProxyUniverse::isSyncronized
 * @return
 */
bool ArbitratingUniverse::isSyncronized() const
{
  auto universe = dominant_();
  if (!universe)
    return false;
  return universe->isSyncronized();
}


/**
 * @brief MergeProxyUniverse::provenance
 * @return
 */
std::shared_ptr<DATA::data_header> ArbitratingUniverse::provenance() const
{
  auto universe = dominant_();
  if (universe)
    return universe->provenance();

  auto prov = std::make_shared<DATA::data_header>();
  prov->universe = expectedUniverse;
  prov->priority = 255; // invalid data
  prov->source_name = "Pending data...";
  return prov;
}


/**
 * @brief MergeProxyUniverse::set
 * @param pdu
 * @param src
 */
void ArbitratingUniverse::set(std::shared_ptr<ACN::DMP::Pdu> pdu,
                             std::shared_ptr<DATA::data_header> src)
{
  if (expectedUniverse != 0 && src->universe != expectedUniverse)
    return;

  if (!sources_.count(*src))
    newProvenance_(*src);

  sources_.at(*src)->set(pdu, src);
}


/**
 * @brief ArbitratingUniverse::synchronize
 * @param sequence_number
 */
void ArbitratingUniverse::synchronize(uint8_t sequence_number)
{
  for ( auto& [_, uni] : sources_)
    uni->synchronize(sequence_number);
}


bool ArbitratingUniverse::isEditable() const {
  return false;
}


uint16_t ArbitratingUniverse::activeSlots() const
{
  auto universe = dominant_();
  if (!universe)
    return 0;
  return universe->activeSlots();
}


/**
 * @brief MergeProxyUniverse::slot
 * @param s
 * @return
 */
uint8_t ArbitratingUniverse::slot(const uint16_t s) const
{
  auto universe = dominant_();
  if (!universe)
    return 0;
  return universe->slot(s);
}


/**
 * @brief MergeProxyUniverse::rxRate
 * @return
 */
double ArbitratingUniverse::rxRate()
{
  auto universe = dominant_();
  if (!universe)
    return 0.0;
  return universe->rxRate();
}


/**
 * @brief MergeProxyUniverse::dominant_
 * @return
 */
Universe* ArbitratingUniverse::dominant_() const
{
  Universe* ret = nullptr;
  for (auto& [_, uni] : sources_)
    {
      if (uni->rxRate() < (DMX::E111_DATA_LOSS_TIMEOUT / 1000.0))
        continue;   // stale universes cannot be dominant
      if (!ret || uni->provenance() > ret->provenance())
        ret = uni;
    }
  return ret;
}


/**
 * @brief MergeProxyUniverse::newSource_
 * @param src
 */
void ArbitratingUniverse::newProvenance_(const DATA::data_header &src)
{
  sources_.emplace(src, new Universe());
  sources_.at(src)->onData(std::bind(&sACN::ArbitratingUniverse::dataChangedNotifier,
                                     this, std::placeholders::_1));
}

} // namespace SACN
