/*
  data.cpp

  Copyright (c) 2020 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#include "dmp/dmp.h"
#include "data.h"

namespace sACN {
namespace DATA {

/**
 * @brief frame_header::iStream
 * @param stream
 */
void data_header::iStream(ACN::PDU::Stream stream)
{
  stream->readString(source_name, 64);
  /// > \cite sACN 6.2.3 E1.31 Data Packet: Priority
  /// >
  /// > No priority outside the range of 0 to 200 shall be transmitted
  /// > on the network.
  *stream >> priority;
  if (priority > 200)
    stream->setstate(std::ios_base::failbit);
  *stream >> sync_address;
  *stream >> sequence_number;
  *stream >> options;
  /// > \cite sACN 6.2.7 E1.31 Data Packet: Universe
  /// >
  /// > Universe values shall be limited to the range 1 to 63999.
  *stream >> universe;
  if ((universe == 0) ||
      (universe >= 64000 && universe != E131_DISCOVERY_UNIVERSE))
    stream->setstate(std::ios_base::failbit);
}


/**
 * @brief data_header::oStream
 * @param stream
 */
void data_header::oStream(ACN::PDU::Stream stream) const
{
  stream->writeString(source_name, 64);
  /// > \cite sACN 6.2.3 E1.31 Data Packet: Priority
  /// >
  /// > No priority outside the range of 0 to 200 shall be transmitted on
  /// > the network.
  *stream << (uint8_t)(priority <= 200 ? priority : 200);
  *stream << sync_address;
  *stream << sequence_number;
  *stream << options;
  *stream << universe;
}


/**
 * @brief Headers are unquely defined by the combination of description,
 *        priority, and universe.
 * @param a
 * @param b
 * @return
 */
bool operator== (const data_header& a, const data_header& b)
{
  return (a.source_name == b.source_name &&
          a.universe == b.universe &&
          a.priority == b.priority);
}


/**
 * @brief For matching universe numbers, lower priorities are lesser.
 * @param a
 * @param b
 * @return
 */
bool operator< (const data_header& a, const data_header& b)
{
  return (a.source_name == b.source_name &&
          a.universe == b.universe &&
          a.priority < b.priority);
}


/**
 * @brief For matching universe numbers, higher priorities are greater.
 * @param a
 * @param b
 * @return
 */
bool operator> (const data_header& a, const data_header& b)
{
  return (a.source_name == b.source_name &&
          a.universe == b.universe &&
          a.priority > b.priority);
}


/**
 * @brief data_options::iStream
 * @param stream
 */
void data_options::iStream(ACN::PDU::Stream stream)
{
  *stream >> byte;
}


/**
 * @brief data_options::oStream
 * @param stream
 */
void data_options::oStream(ACN::PDU::Stream stream) const
{
  *stream << byte;
}


/**
 * @brief Pdu::Pdu
 * @param stream
 */
Pdu::Pdu()
  : ACN::PDU::Pdu(4) // vectors are 4 octets
{
}


/**
 * @brief Pdu::iStream
 * @param stream
 */
void Pdu::iStream(ACN::PDU::Stream stream)
{
  ACN::PDU::Pdu::iStream(stream);         // flags, length, and vector
  createHeader<data_header>();            // header
  createDataBlock<ACN::DMP::Pdu>();       // data
}

} // DATA
} // SACN
