/*
  universe.h

  Copyright (c) 2020 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/
#pragma once

#include "data.h"
#include "dmp/dmp.h"
#include "dmx/universe.h"

#include <cstdint>
#include <condition_variable>
#include <memory>
#include <thread>
#include <vector>

namespace sACN {

class Source; ///< forward declare from "sacn/source.h"

/**
 * @brief \cite sACN 3.2 Universe
 *
 * > A set of up to 512 data slots identified by universe number.
 * > Note: In E1.31 there may be multiple sources for a universe. See also: Slot.
 */
class Universe
    : public DMX::Universe
{
public:
  Universe(Source* = nullptr);
  ~Universe();

  virtual void set(ACN::PDU::Message<ACN::DMP::Pdu>, std::shared_ptr<DATA::data_header>);

  virtual std::shared_ptr<DATA::data_header> provenance() const;
  virtual void setProvenance(std::shared_ptr<DATA::data_header>);

  virtual bool isSyncronized() const;
  virtual void synchronize(uint8_t = 0);

  virtual bool isEditable() const;
  virtual uint16_t activeSlots() const;

  // DMX::Universe overrides
  void setValue (const uint16_t address, const uint8_t value) override;
  void setValue (const uint16_t start, const uint16_t footprint,
                 const uint8_t* data) override;

  /**
   * @brief destination IP address
   *
   * Setting type to ACN::SDT::SDT_ADDR_NULL will send on the multicast group.
   */
  ipAddress destination;



protected:
   void sACNsend() const;

   /**
    * @brief \cite sACN 3.7 Active Data Slots
    *
    * > When translating ANSI E1.11 DMX512-A \cite DMX to E1.31, the active data
    * > slots are defined as ranging from data slot 1 to the maximum data slot in
    * > the most recently received packet with the corresponding START Code.
    * >
    * > Devices originating E1.31 shall define their active data slots using
    * > the DMP First Property Address and DMP Property Count fields shown in
    * > Table 4-1.
    */
    uint16_t active_data_slots;

private:
  std::shared_ptr<DATA::data_header> provenance_;
  std::vector<uint8_t> * sync_data_ = nullptr;
  Source* source_;

  /// > \cite sACN 6.3.2 E1.31 Synchronization Packet: Sequence Number
  /// >
  /// > Sources shall maintain a sequence for each universe they transmit.
  /// > The sequence number for a universe shall be incremented by one for
  /// > every packet sent on that universe. There is no implied relationship
  /// > between the sequence number of an E1.31 Synchronization Packet and
  /// > the sequence number of an E1.31 Data Packet on that same universe.
  uint8_t sync_sequence_ = 0;

  std::condition_variable_any tx_request_;
  std::mutex tx_control_mutex_;
  bool tx_enable_ = true;
  std::thread tx_worker_;
  void tx_loop_();
};

} // SACN namespace
